module Datapath where

import List
import STATE
import ErrM
import SymbVal

type Datapath = ([Node], [Connection])

type NodeName = String

data Node = 
 Function
 {
   nodename :: NodeName,
   func :: SymbFunc
 }
 | JoinPoint
 {
   nodename :: NodeName
 }
 | Input
 {
   nodename :: NodeName
 }
 | Output
 {
   nodename :: NodeName
 }
 | Delay
 {
   nodename :: NodeName,
   value :: SymbVal
 }
instance Eq Node where
 n == m = (nodename n) == (nodename m)
instance Show Node where
 show (Delay s v) = "|" ++ s ++ "|" ++ (show v) ++ ">"
 show (Output s) = "//" ++ s
 show (Input s) = s ++ "\\\\"
 show (JoinPoint s) = s ++ "}"
 show (Function s f) =  "{" ++ s ++ ":" ++ (show f) ++ "}"

type InputPort = (NodeName, Int)
type Connection = (NodeName, InputPort)
     --           (from,(to, port#))
                   

type Result = STATE Datapath

--Node and Edge creation functions

addEdge :: Connection -> Result ()
addEdge e@(f, (t,n)) =
 do
   dp@(ns, es) <- readSTATE
   if elem e es
     then fail "Tried to add a duplicate connection!"
     else do
       to <- lookupNodeM t
       from <- lookupNodeM f
       case (from,to) of --for modular DPs
         (Output _, Input _) -> do --replace with a direct connection
           let prv = getPrevNode_n dp (f, 0)
           let nexts = getNextNodes_n dp t
           remNode t
           remNode f
           let zips = zip (replicate (length nexts) prv) nexts
           mapM connect_n zips
           return ()
         (_, Input _) -> do
           let nexts = getNextCons dp t
           mapM (reanchor f) nexts
           remNode t
         (Output _, _) -> do
           let prv = getPrevCon_n dp (f, 0)
           repoint (t,n) prv
           remNode f
         z -> writeSTATE (ns, e:es) --normal add

addEdges es = 
 do
   mapM addEdge es
   return ()

connect :: (NodeName, NodeName) -> Result ()
connect (f, t) =
 do
   t' <- lookupNodeM t
   case t' of
     (Function _ func) -> if nParams func > 1 
        then fail "Use connect_n to connect to multi-parameter functions."
        else addEdge (f, (t, 0))
     z -> addEdge (f, (t, 0))

connect_n :: (NodeName, InputPort) -> Result ()
connect_n (f, t) =
 do
   addEdge (f, t)

connections :: [(NodeName, NodeName)] -> Result ()
connections zs =
 do
   mapM connect zs
   return ()

connections_n :: [(NodeName,InputPort)] -> Result ()
connections_n zs =
 do
   mapM connect_n zs
   return ()

--connects a node to all elements of the list
connectTo :: NodeName -> [NodeName] -> Result ()
connectTo from tos = 
 do
   let zips = zip (replicate (length tos) from) tos
   connections zips

connectTo_n :: NodeName -> [InputPort] -> Result ()
connectTo_n from tos = 
 do
   let zips = zip (replicate (length tos) from) tos
   connections_n zips
 
addNode :: Node -> Result ()
addNode n =
 do 
   (ns, es) <- readSTATE
   if elem n ns
     then fail ("Non-unique node name: " ++ (nodename n))
     else writeSTATE (n:ns, es)

addNodes ns = 
 do 
   mapM addNode ns
   return ()

input :: NodeName -> Result ()
input = addNode . Input

inputs :: [NodeName] -> Result ()
inputs ns = 
 do
    let is = map Input ns
    addNodes is

output :: NodeName -> Result ()
output = addNode . Output
    
outputs :: [NodeName] -> Result ()
outputs ns = 
 do
    let is = map Output ns
    addNodes is

delay :: SymbVal -> NodeName -> Result ()
delay v n = addNode (Delay n v)

delays :: SymbVal -> [NodeName] -> Result ()
delays v ns =
 do
   let is = map (flip Delay v) ns
   addNodes is

--N delays with the same value, linked in a line
delayN :: Int -> SymbVal -> NodeName -> Result ()
delayN 0 _ _ = return ()
delayN 1 v nm =
 do
   delay v (nm ++ "_0")
delayN n v nm =
 do
   let ns = map show (range (0, n - 1))
   let nms = map ((nm ++ "_") ++) ns
   delays v nms
   linkUp (reverse nms)

--Creates a linked line of delays with the given initial values
delayList :: [SymbVal] -> NodeName -> Result ()
delayList [] nm = return ()
delayList [v] nm =
 do
   delay v (nm ++ "_0")
delayList vs nm =
 do
   let ns = map show (range (0, (length vs) - 1))
   let nms = map ((nm++"_")++) ns
   let zips = zip vs nms
   mapM (uncurry delay) zips
   linkUp (reverse nms)
   
function :: SymbFunc -> NodeName -> Result ()
function f n = addNode (Function n f)

functions :: SymbFunc -> [NodeName] -> Result ()
functions f ns =
 do
   let is = map (flip Function f) ns
   addNodes is

joinPoint :: NodeName -> Result ()
joinPoint n = addNode (JoinPoint n)

joinPoints :: [NodeName] -> Result ()
joinPoints ns =
 do
   let is = map (JoinPoint) ns
   addNodes is   

remEdge :: Connection -> Result ()
remEdge e = 
 do
   (ns, es) <- readSTATE
   writeSTATE (ns, es \\ [e])

remEdges es = 
 do
   mapM remEdge es
   return ()

disconnect :: NodeName -> NodeName -> Result ()
disconnect f t =
 do
   t' <- lookupNodeM t
   case t' of
     (Function _ func) -> if nParams func > 1 
        then fail "Use disconnect_n to disconnect from multi-parameter functions."
        else remEdge (f, (t, 0))
     z -> remEdge (f, (t, 0))


disconnect_n :: NodeName -> InputPort -> Result ()
disconnect_n f t =
 do
   remEdge (f, t)

remNode :: NodeName -> Result ()
remNode node =
 do
   n <- lookupNodeM node
   dp <- readSTATE
   let cs = getEdges dp (nodename n)
   remEdges cs
   (ns, es) <- readSTATE
   writeSTATE (ns \\ [n], es)

remNodes ns = 
 do
   mapM remNode ns
   return ()

--Non-monadic data-gathering helper functions

getEdges :: Datapath -> NodeName -> [Connection]
getEdges dp n = next ++ prvs
 where
   prvs = getPrevCons dp n
   next = getNextCons dp n

getNextCons :: Datapath -> NodeName -> [Connection]
getNextCons (_, es) n = nexts
 where
  nexts = filter (isPrev n) es

getNextNodes :: Datapath -> NodeName -> [NodeName]
getNextNodes dp n = map (fst . snd) cs
 where
   cs = getNextCons dp n

getNextNodes_n :: Datapath -> NodeName -> [InputPort]
getNextNodes_n dp n = map snd (getNextCons dp n)

getPrevCons :: Datapath -> NodeName -> [Connection]
getPrevCons dp@(ns, es) node =
 case lookupNode dp node of 
    (Function f func) -> concatMap (getPrev' f es) (range (0, (nParams func) - 1))
    (JoinPoint f) -> concatMap (getPrev' f es) [0, 1]
    n -> filter (isNext (node,0)) es
  where
   getPrev' f es n = filter (isNext (f,n)) es

getPrevNodes :: Datapath -> NodeName -> [NodeName]
getPrevNodes dp node = map fst cs
 where
  cs = getPrevCons dp node

getPrevCon_n :: Datapath -> InputPort -> Connection
getPrevCon_n dp@(ns, es) (node, n) =
 case prvs of
    [] -> error ("No previous node for node " ++ node ++ ", input " ++ (show n))
    (x:[]) -> x
    xs -> error ("Somehow there is more than one node connected to " ++ node ++ ", input " ++ (show n))
  where
   prvs = filter (isNext (node,n)) es

getPrevCons_n :: Datapath -> InputPort -> [Connection]
getPrevCons_n (_, es) (node, n) = filter (isNext (node,n)) es
   
getPrevNode_n :: Datapath -> InputPort -> NodeName
getPrevNode_n dp p = fst (getPrevCon_n dp p)

getPrevNodes_n :: Datapath -> InputPort -> [NodeName]
getPrevNodes_n (_, es) (node, n) = map fst prvs
  where
    prvs = filter (isNext (node,n)) es
    
lookupNode :: Datapath -> NodeName -> Node
lookupNode (ns, _) s =
   case find (\x -> nodename x == s) ns of
     Just x -> x
     Nothing -> error ("Unable to find a node named " ++ s)

lookupNodeM :: NodeName -> Result Node
lookupNodeM s =
 do
   (ns, _) <- readSTATE
   case find (\x -> nodename x == s) ns of
     Just x -> return x
     Nothing -> fail ("Unable to find a node named " ++ s)
     
  
isNext :: InputPort -> Connection -> Bool
isNext (s,k) (f,(t,n)) = t == s && k == n

isPrev :: NodeName -> Connection -> Bool
isPrev s (f,(t,n)) = f == s 
   
isConnected :: Datapath -> InputPort -> Bool
isConnected (_, es) c = prvs /= []
 where
   prvs = filter (isNext c) es

isDelay :: Node -> Bool
isDelay (Delay _ _) = True
isDelay _ = False

isInput :: Node -> Bool
isInput (Input _) = True
isInput _ = False

isOutput :: Node -> Bool
isOutput (Output _) = True
isOutput _ = False

isFunction :: Node -> Bool
isFunction (Function _ _) = True
isFunction _ = False

isJoinPoint :: Node -> Bool
isJoinPoint (JoinPoint _) = True
isJoinPoint _ = False

getNodes :: Datapath -> [NodeName]
getNodes (ns, _) = map nodename ns

getDelays :: Datapath -> [Node]
getDelays (ns, _) = filter isDelay ns

getInputs :: Datapath -> [Node]
getInputs (ns, _) = filter isInput ns

getOutputs :: Datapath -> [Node]
getOutputs (ns, _) = filter isOutput ns

getFunctions :: Datapath -> [Node]
getFunctions (ns, _) = filter isOutput ns

getJoinPoints :: Datapath -> [Node]
getJoinPoints (ns, _) = filter isJoinPoint ns

getInitialState :: Datapath -> [(NodeName, SymbVal)]
getInitialState dp = map getIV (getDelays dp)
 where
  getIV (Delay n v) = (n, v)

--Returns all non-JP nodes with more than one incoming connection per port
--Used to add implicit JPs
getMultipleCons :: Datapath -> [InputPort]
getMultipleCons (ns, es) = concatMap hasMCs ns
 where
  hasMCs (JoinPoint s) = []
  hasMCs (Function s f) = concatMap checkPrev (zip (replicate k s) (range (0, k - 1)))
   where
    k = nParams f
  hasMCs n = checkPrev (nodename n, 0)
  checkPrev p = if length (getPrevCons_n (ns, es) p) > 1 
                then [p]
                else []

--Datapath altering helper functions

--alters the destination of a connection
repoint :: InputPort -> Connection -> Result ()
repoint to e@(from, t) =
 do
   (ns,es) <- readSTATE
   let es' = es \\ [e]
   writeSTATE (ns, (from, to):es')

--alters the origin of a connection
reanchor :: NodeName-> Connection -> Result ()
reanchor from e@(f, to) =
 do
   (ns,es) <- readSTATE
   let es' = es \\ [e]
   writeSTATE (ns, (from, to):es')

--creates Delay D and JoinPoint D_load, linked
register :: SymbVal -> NodeName -> Result ()
register v r =
 do
   let load = r ++ "_load"
   addNodes [JoinPoint load, Delay r v]
   addEdges [(load,(r,0)), (r, (load, 0))]

--adds a register before the specified port
--incoming connections are re-pointed       
registerBefore :: SymbVal -> NodeName -> InputPort -> Result ()
registerBefore v r (to, n) =
 do
   let load = r ++ "_load"
   dp <- readSTATE
   let cs = getPrevCons_n dp (to, n)
   addNodes [JoinPoint load, Delay r v]
   addEdges [(load,(r,0)), (r, (load, 0)), (r, (to, n))]
   mapM_ (repoint (load, 0)) cs

--adds a register after the specified node
--outgoing connections are re-anchored
registerAfter :: SymbVal -> NodeName -> NodeName -> Result ()
registerAfter v r from =
 do
   let load = r ++ "_load"
   dp <- readSTATE
   let cs = getNextCons dp from
   addNodes [JoinPoint load, Delay r v]
   addEdges [(load,(r,0)), (r, (load, 0)), (from, (r, 0))]
   mapM_ (reanchor r) cs
   
--Links up the nodes in the list
linkUp :: [NodeName] -> Result ()
linkUp [] = return ()
linkUp [x] = return ()
linkUp (x:y:xs) =
 do
   addEdge (x,(y,0))
   linkUp (y:xs)

linkUp_n :: [InputPort] -> Result ()
linkUp_n [] = return ()
linkUp_n [x] = return ()
linkUp_n ((x,_):(y,n):xs) =
 do
   addEdge (x,(y,n))
   linkUp_n ((y,n):xs)

--Creates a Delay linked to itself
constant :: SymbVal -> NodeName -> Result ()
constant v n =
 do
   addNode (Delay n v)
   addEdge (n, (n, 0))

--Creates a list of delays linked to the beginning
constantList :: [SymbVal] -> NodeName -> Result ()
constantList [] _ = return ()
constantList [x] s = constant x (s ++ "_0")
constantList vs nm = 
 do
   let nm' = nm ++ "_"
   let k = (length vs) - 1
   let ns = map show (range (0,k))
   let nms = map (nm'++) ns
   let zips = zip nms vs
   let ds = map (uncurry Delay) zips
   addNodes ds
   linkUp (reverse nms)
   addEdge (nm' ++ "0", (nm' ++ (show k), 0))

--Creates two lists, linked, the second of which repeats
oneTimeList :: ([SymbVal], [SymbVal]) -> NodeName -> Result ()
oneTimeList (is, fs) nm =
 do
   constantList fs (nm ++ "_init")
   delayList is nm
   connect ((nm ++ "_init_0"), (nm ++ "_" ++ (show ((length is) - 1))))

--renames nodes, edges for modular datapaths
prefix :: String -> String -> String
prefix s n = s ++ ":" ++ n

nodePrefix :: String -> Node -> Node
nodePrefix s (Delay nm (Var v)) = (Delay (prefix s nm) (Var (prefix s v)))
nodePrefix s (Delay nm (VarIx v x)) = (Delay (prefix s nm) (VarIx (prefix s v) x))
nodePrefix s n = n {nodename = (prefix s (nodename n))}

edgePrefix :: String -> Connection -> Connection
edgePrefix s (f, (t, n)) = (prefix s f, (prefix s t, n))

--Embeds a previously defined DP
addDP :: Datapath -> String -> Result ()
addDP (ns,es) nm = 
 do --rename all units, fix connections, add.
   let ns' = map (nodePrefix nm) ns
   let es' = map (edgePrefix nm) es
   mapM addNode ns'
   mapM addEdge es'
   return ()

--Adds one JP and repoints incoming connections to it.
addJoinPoint :: InputPort -> Result ()
addJoinPoint (to, n) =
 do
   dp <- readSTATE
   let prvs = getPrevCons_n dp (to, n)
   let s = "into_(" ++ to ++ "," ++ (show n) ++ ")"
   addNode (JoinPoint s)
   mapM (repoint (s, 0)) prvs
   connect_n (s, (to, n))
   return ()

--Adds implicit JPs
addJoinPoints :: Result ()
addJoinPoints =
 do
   dp <- readSTATE
   let ns = getMultipleCons dp
   mapM addJoinPoint ns
   return ()
   
makeDatapath dp = case appSTATE (dp >> checkDatapath) ([],[]) of Ok (x, y) -> x

checkDatapath :: Result Datapath
checkDatapath = 
 do
   addJoinPoints
   checkOutputs
   checkCons
   checkGraph
   dp <- readSTATE
   return dp

 
checkOutputs :: Result ()
checkOutputs =
 do
  dp <- readSTATE
  case getOutputs dp of
   [] -> error "Datapath must include at least one Output node!"
   _ -> return ()
     
--All ports must be connected, except JP control ports
--More than one edge may originate from a output port,
--but multiple edges may not be connected to one input port. 
checkCons :: Result ()
checkCons =
 do
   (ns, es) <- readSTATE
   mapM_ checkCon ns

checkCon :: Node -> Result ()
checkCon (Input nm) =
  do
   dp <- readSTATE
   case getNextCons dp nm of
    [] -> error ("No outgoing connection for Input " ++ nm ++ "!")
    _ -> return ()
checkCon (Output nm) =
  do
   dp <- readSTATE
   case getPrevCon_n dp (nm, 0) of --this function does the input checking
    c -> return ()
checkCon (Delay nm val) =
  do
   dp <- readSTATE
   case getPrevCon_n dp (nm, 0) of
    c -> case getNextCons dp nm of
     [] -> error ("No outgoing connection for Delay " ++ nm ++ "!")
     _ -> return ()
checkCon (JoinPoint nm) =
  do
   dp <- readSTATE
   case getPrevCons_n dp (nm, 0) of
    [] -> error ("No incoming connections to Join Point " ++ nm ++ " input port 0!")
    c -> case getNextCons dp nm of
     [] -> error ("No outgoing connection for Join Point " ++ nm ++ "!")
     _ -> return ()
checkCon (Function nm f) =
  do
   dp <- readSTATE
   let k = nParams f
   let zips = zip (replicate k nm) (range (0, k - 1))
   mapM (checkFun dp) zips
   case getNextCons dp nm of
     [] -> error ("No outgoing connection for Function " ++ nm ++ "!")
     _ -> return ()
  where
   checkFun dp port = case getPrevCon_n dp port of
                        c -> return ()


--check that the datapath is weakly connected,
--IE no disjoint components. Also checks for combinational loops.
--A straightforward fixed-point iteration from output nodes
checkGraph :: Result ()
checkGraph = 
 do
   dp <- readSTATE
   let outs = map nodename (getOutputs dp) --guaranteed non-empty
   coverage <- expand outs
   if coverage `setEquals` (getNodes dp)
     then return ()
     else error ("Disjoint component found including nodes: " ++
	        (unwords coverage))

expand :: [NodeName] -> Result [NodeName]
expand ns =
 do
  dp <- readSTATE
  let prvs = concatMap (getPrevNodes dp) ns
  let is = intersect prvs ns
  mapM checkLoops is
  let nexts = union ns prvs
  if nexts `setEquals` ns
    then return nexts
    else expand nexts

--Checks for a possible combinational loop.
checkLoops :: NodeName -> Result ()
checkLoops n =
 do
   dp <- readSTATE
   node <- lookupNodeM n
   case node of
    (Delay _ _) -> return ()
    _ -> do
      let prvs = getPrevNodes dp n
      mapM (checkPath n []) prvs
      return ()

--All loop paths must include at least one Delay node
checkPath :: NodeName -> [NodeName] -> NodeName -> Result ()
checkPath orig visited n = 
 do
  dp <- readSTATE
  node <- lookupNodeM n
  case node of
    (Delay _ _) -> return ()
    _ -> if elem n visited
      then return ()
      else do
        if orig == n
          then error ("Combinational Loop found from node " ++ orig ++ " to " ++ n ++ "!")
          else do
            let prvs = getPrevNodes dp n
            mapM_ (checkPath orig (n:visited)) prvs
      
    

setEquals :: Ord a => [a] -> [a] -> Bool
setEquals x y = sort x == sort y
