
import CoreLoom

import STATE
import ErrM

--A simple R = f(X) datapath
--With feedback and externally controlled registers
dp0 :: Datapath
dp0 = makeDatapath dp
 where
  dp = do
   input "X"
   function (SymbFunc "f" 1) "F"
   output "R"
   connections [("X", "F"), ("F", "R"), ("F", "F")]
   registerBefore open "R1" ("F", 0)
   registerBefore open "R2" ("R", 0)

--Straight through the dp
dp0_flat :: Config
dp0_flat = 
  (["R1_load",      "R2_load"],
   [[Var "X",         open],
    [open,            Var "F"]])

--an example of looping with reduced throughput
dp0_loop :: Config
dp0_loop = 
  (["R1_load",      "R2_load"],
   [[Var "X",        open],
    [Var "F",        open],
    [Var "F",        open],
    [open,           Var "F"]])

-- R = f(X, Y) with feedback, registers
-- Extra latency added
dp1 :: Datapath
dp1 = makeDatapath dp
 where
  dp = do
   inputs ["X", "Y"]
   function (SymbFunc "f" 2) "F" 
   output "R"
   delay open "D1"
   delay open "D2"
   connections_n [("X", ("F", 0)), ("Y", ("F", 1)), ("D2", ("F", 0)), ("D2", ("F", 1))]
   connections [("F", "D1"), ("D1", "D2"), ("D2", "R"),("X", "R"), ("Y", "R")]
   registerBefore open "R0" ("F", 0)
   registerBefore open "R1" ("F", 1)
   registerBefore open "R2" ("R", 0)

--Straight through
dp1_flat :: Config
dp1_flat = 
  (["R0_load",     "R1_load",      "R2_load"],
   [[Var "X",       Var "Y",        open],
   [open,           open,           open],
   [open,           open,           open],
   [open,           open,           Var "D2"]])

--An example of how scheduling improves throughput.     
cf_dp1_sched :: Config
cf_dp1_sched = 
  (["R0_load",      "R1_load",      "R2_load"],
  [[Var "X",        Var "Y",        open],
  [open,            open,           open],
  [open,            open,           open],
  [Var "X",         Var "D2",       open],
  [open,            open,           open],
  [open,            open,           open],
  [open,            open,           Var "D2"]])


--Demonstrates implicit join points
--  Main> configInfo dp2
--   [("into_(R,0)",["A","B","C","D","E"])]
dp2 :: Datapath
dp2 = makeDatapath dp
 where
  dp = do
   inputs ["A","B","C","D", "E"]
   outputs ["R"]
   connections [("A","R"),("B","R"), ("C", "R"), ("D", "R"), ("E", "R")]

--full adder example
fadd :: Datapath
fadd = makeDatapath dp
 where
  dp = do
   inputs ["X", "Y"]
   outputs ["R", "C"]
   addNode (Function "ADD" XOR)
   addNode (Function "CRY" AND)
   connections_n [("X",("ADD", 0)),("Y", ("ADD", 1)),("X", ("CRY", 0)), ("Y", ("CRY", 1))]
   connections [("ADD","R"), ("CRY", "C")]

--demonstrates embedded modules 
embed_test :: Datapath
embed_test = makeDatapath dp
 where
  dp = do
   inputs ["X", "Y", "Z", "W"]
   outputs ["R"]
   addNode (Function "F" XOR)
   addNode (Function "F2" XOR)
   addNode (Function "F3" XOR)
   addDP fadd "ADD1"
   addDP fadd "ADD2"
   connections [("X", "ADD1:X"),("Y", "ADD1:Y"), ("Z", "ADD2:X"),("W", "ADD2:Y")]
   connections_n [("ADD1:R", ("F", 0)), ("ADD1:C", ("F2", 0))]
   connections_n [("ADD2:R", ("F", 1)), ("ADD2:C", ("F2", 1))]
   connections_n [("F", ("F3", 0)), ("F2", ("F3", 1))]
   connections [("F3", "R")]

--Calculates Dot Product R=AC+BD inflexibly
dotP :: Datapath 
dotP = makeDatapath dp
 where
  dp = do
    inputs ["A","B","C","D"]
    delays open ["d_0", "d_1", "d_2"]
    functions MUL ["MUL_0", "MUL_1"]
    function ADD "ADD_0"
    output "R"
    connections_n [("A", ("MUL_0",0)), ("C", ("MUL_0",1))]
    connections_n [("B" ,("MUL_1",0)), ("D", ("MUL_1",1))]
    connections [("MUL_0", "d_0"), ("MUL_1", "d_1")]
    connections_n [("d_0", ("ADD_0", 0)), ("d_1", ("ADD_0", 1))]
    connections [("ADD_0", "d_2"), ("d_2", "R")]

--Dot Product with added flexibility
flex_dotP :: Datapath 
flex_dotP = makeDatapath dp
 where
  dp = do
    inputs ["A","B","C","D"]
    delays open ["d_0", "d_1", "d_2"]
    functions MUL ["MUL_0", "MUL_1"]
    function ADD "ADD_0"
    output "R"
    connections_n [("A", ("MUL_0",0)), ("C", ("MUL_0",1))]
    connections_n [("B" ,("MUL_1",0)), ("D", ("MUL_1",1))]
    connections [("MUL_0", "d_0"), ("MUL_1", "d_1")]
    connections_n [("d_0", ("MUL_0",1)), ("d_1", ("MUL_1", 0))]
    connections_n [("d_0", ("ADD_0", 0)), ("d_1", ("ADD_0", 1))]
    connections [("ADD_0", "d_2"), ("d_2", "R")]
    connectTo_n "d_2" [("MUL_0", 0), ("MUL_1", 1)]

--Flat AC+BD
dotP_dotP :: Config
dotP_dotP = 
  (["into_(MUL_0,0)","into_(MUL_0,1)","into_(MUL_1,0)","into_(MUL_1,1)"],
   [[Var "A",         Var "C",         Var "B",         Var "D"]])

--An example of feedback
dotP_back :: Config
dotP_back = 
  (["into_(MUL_0,0)","into_(MUL_0,1)","into_(MUL_1,0)","into_(MUL_1,1)"],
   [[Var "A",         Var "C",         Var "B",         Var "D"],
    [open,            open,            open,            open],
    [Var "d_2",       Var "C",         Var "B",         Var "D"]])

--An example of a "stuck" configuration
dotP_stuck :: Config
dotP_stuck = (["into_(MUL_0,0)","into_(MUL_0,1)","into_(MUL_1,0)","into_(MUL_1,1)"],
             [[Var "d_2",   Var "d_0",   Var "d_1",   Var "d_2"]])


--Dot Product with registers
reg_dotP :: Datapath 
reg_dotP = makeDatapath dp
 where
  dp = do
    inputs ["A","B","C","D"]
    functions MUL ["MUL_0", "MUL_1"]
    function ADD "ADD_0"
    output "R"
    connections_n [("A", ("MUL_0",0)), ("C", ("MUL_0",1))]
    connections_n [("B" ,("MUL_1",0)), ("D", ("MUL_1",1))]
    connections_n [("MUL_0", ("ADD_0", 0)), ("MUL_1", ("ADD_0", 1))]
    connections_n [("MUL_0", ("MUL_0",1)), ("MUL_1", ("MUL_1", 0))]
    connectTo_n "ADD_0" [("R", 0), ("MUL_0", 0), ("MUL_1", 1)]
    connectTo_n "ADD_0" [("ADD_0", 0), ("ADD_0", 1)]
    registerBefore open "d_0" ("MUL_0", 0)
    registerBefore open "d_1" ("MUL_0", 1)
    registerBefore open "d_2" ("MUL_1", 0)
    registerBefore open "d_3" ("MUL_1", 1)
    registerBefore open "d_4" ("ADD_0", 0)
    registerBefore open "d_5" ("ADD_0", 1)
    registerBefore open "d_6" ("R", 0)
    
--AC+BD    
reg_dotP_flat :: Config
reg_dotP_flat = 
  (["d_0_load", "d_1_load", "d_2_load", "d_3_load", "d_4_load", "d_5_load", "d_6_load"],
  [[Var "A",     Var "C",    Var "B",    Var "D",    open,       open,       open],
   [open,        open,       open,       open,      Var "MUL_0", Var "MUL_1",open],
   [open,        open,       open,       open,      open,        open,       Var "ADD_0"]])

--demonstrates repeated scheduling
reg_dotP_rep :: Config
reg_dotP_rep = 
  (["d_0_load", "d_1_load", "d_2_load", "d_3_load", "d_4_load", "d_5_load", "d_6_load"],
  [[Var "A",     Var "C",    Var "B",    Var "D",    open,       open,       open],
   [Var "A",     Var "MUL_0",Var "MUL_1",Var "D",   Var "MUL_0", Var "MUL_1",open],
   [Var "ADD_0", Var "C",    Var "B",   Var "ADD_0",Var "MUL_0", Var "MUL_1",Var "ADD_0"],
   [open,        open,       open,      open,       Var "MUL_0", Var "MUL_1",Var "ADD_0"],
   [open,        open,       open,      open,       open,        open,       Var "ADD_0"]])

--demonstrates overlapping scheduling
reg_dotP_over :: Config
reg_dotP_over = 
  (["d_0_load", "d_1_load", "d_2_load", "d_3_load", "d_4_load", "d_5_load", "d_6_load"],
  [[Var "A",     Var "C",    Var "B",    Var "D",    open,       open,       open],
   [open,        open,       open,       open,       Var "MUL_0",Var "MUL_1",open],
   [Var "ADD_0", Var "C",    Var "B",    Var "D",    open,       open,       open],
   [open,        open,       open,       open,       Var "MUL_0",Var "MUL_1",open],
   [open,        open,       open,       open,       open,       open,       Var "ADD_0"]])

--A configuration which is impossible to schedule
reg_dotP_nosched :: Config
reg_dotP_nosched = 
  (["d_0_load", "d_1_load", "d_2_load", "d_3_load", "d_4_load", "d_5_load", "d_6_load"],
   [[Var "A",    Var "C",    Var "B",    Var "D",    open,       open,       open],
    [open,       open,       open,       open,       Var "MUL_0",Var "MUL_1",open],
    [Var "ADD_0",Var "C",    Var "B",    Var "ADD_0",open,       open,       open],
    [open,       open,       open,       open,       Var "MUL_0",Var "MUL_1",open],
    [Var "ADD_0",Var "MUL_0",Var "MUL_1",Var "ADD_0",Var "d_4",  Var "d_5",  Var "ADD_0"]])

--An example of a Black-Box datapath
--Represents an inflexible encryptor, unpipelined
--latency = 4, TP = 1/4
exampleF :: Datapath
exampleF = makeDatapath dp
 where
  dp = do
   inputs ["Key", "Data"]
   outputs ["Cipher"]
   function (SymbFunc "encrypt" 2) "ENC"
   connections_n [("Key", ("ENC", 0)), ("Data", ("ENC", 1))]
   connect ("ENC", "Cipher")
   registerBefore open "R0" ("Cipher", 0)
   registerBefore open "R1" ("Cipher", 0)
   constantList [Var "ENC", Var "R0", Var "R0", Var "R0"] "OneList"
   constantList [Var "R1", Var "R1", Var "R1", Var "R0"] "TwoList"
   connect_n ("OneList_0", ("R0_load", 1))
   connect_n ("TwoList_0", ("R1_load", 1))

--An example of a black-box datapath
--represents an arbitrary usage pattern
exampleG :: Datapath
exampleG = makeDatapath dp
 where
  dp = do
   inputs ["Key", "Data"]
   outputs ["Cipher"]
   function (SymbFunc "encrypt" 2) "ENC"
   connections_n [("Key", ("ENC", 0)), ("Data", ("ENC", 1))]
   connect ("ENC", "Cipher")
   registerBefore open "R0" ("Cipher", 0)
   registerBefore open "R1" ("Cipher", 0)
   registerBefore open "R2" ("Cipher", 0)
   constantList [Var "ENC", Var "ENC", Var "R0", Var "R0"] "OneList"
   constantList [Var "R1", Var "R0", Var "R0", Var "R1"] "TwoList"
   constantList [open, open, Var "R1", Var "R1"] "ThreeList"
   connect_n ("OneList_0", ("R0_load", 1))
   connect_n ("TwoList_0", ("R1_load", 1))
   connect_n ("ThreeList_0", ("R2_load", 1))

--A model of OpenCore.org's AES core encryptor
--Latency = 12 TP = 1/12
aes_enc :: Datapath
aes_enc = makeDatapath dp
 where
  dp = do
   inputs ["key", "text_in"]
   outputs ["text_out"]
   function (SymbFunc "KeyExpansion" 1) "EXP"
   function (SymbFunc "InitialPermutation" 2) "INIT"
   function (SymbFunc "RoundPermutation" 2) "ROUND"
   function (SymbFunc "FinalPermutation" 2) "FINAL"
   register open "R0"
   register open "R1"
   delay open "R2"
   register open "R3"
   joinPoint "JP"
   connections [("key", "R0_load"), ("R0", "EXP"), ("text_in", "R1_load")]
   connectTo_n "EXP" [("INIT", 0), ("ROUND", 0), ("FINAL", 0)]
   connections_n [("R1", ("INIT", 1)), ("INIT", ("JP", 0)), ("JP", ("ROUND", 1))]
   connections_n [("ROUND", ("R2", 0)), ("R2", ("JP", 0)), ("R2", ("FINAL", 1))]
   connections_n [("FINAL", ("R3_load", 0)), ("R3", ("text_out", 0))]
   constantList ((Var "key") : (replicate 11 (Var "R0"))) "ZeroList"
   constantList ((Var "text_in") : (replicate 11 (Var "R1"))) "OneList"
   constantList ([Var "R2", Var "INIT"] ++ (replicate 10 (Var "R2"))) "TwoList"
   constantList ((replicate 11 open) ++ [Var "FINAL"]) "ThreeList"
   connect_n ("ZeroList_0", ("R0_load", 1))
   connect_n ("OneList_0", ("R1_load", 1))
   connect_n ("TwoList_0", ("JP", 1))
   connect_n ("ThreeList_0", ("R3_load", 1))

--A model of OpenCore.org's decryptor
--The key is loaded before the text
--Latency = 24, TP = 1/24
aes_dec :: Datapath
aes_dec = makeDatapath dp
 where
  dp = do
   inputs ["key", "text_in"]
   outputs ["text_out"]
   function (SymbFunc "KeyExpansion" 1) "EXP"
   function (SymbFunc "KeyReversal" 1) "REV"
   function (SymbFunc "InitialInverse" 2) "INIT"
   function (SymbFunc "RoundInverse" 2) "ROUND"
   function (SymbFunc "FinalInverse" 2) "FINAL"
   register open "R0"
   register open "R1"
   delay open "R2"
   register open "R3"
   joinPoint "JP"
   connections [("key", "R0_load"), ("R0", "EXP"), ("EXP", "REV"), ("text_in", "R1_load")]
   connectTo_n "REV" [("INIT", 0), ("ROUND", 0), ("FINAL", 0)]
   connections_n [("R1", ("INIT", 1)), ("INIT", ("JP", 0)), ("JP", ("ROUND", 1))]
   connections_n [("ROUND", ("R2", 0)), ("R2", ("JP", 0)), ("R2", ("FINAL", 1))]
   connections_n [("FINAL", ("R3_load", 0)), ("R3", ("text_out", 0))]
   oneTimeList ([Var "key"], [Var "R0"]) "ZeroList"
   constantList ((replicate 12 (Var "R1")) ++ [Var "text_in"] ++ (replicate 11 (Var "R1"))) "OneList"
   constantList ((replicate 13 (Var "R2")) ++ [Var "INIT"] ++ (replicate 10 (Var "R2"))) "TwoList"
   constantList ((replicate 23 open) ++ [Var "FINAL"]) "ThreeList"
   connect_n ("ZeroList_0", ("R0_load", 1))
   connect_n ("OneList_0", ("R1_load", 1))
   connect_n ("TwoList_0", ("JP", 1))
   connect_n ("ThreeList_0", ("R3_load", 1))

--A combination of the encryption and decryption cores

aes_comb :: Datapath
aes_comb = makeDatapath dp
 where
  dp = do
   inputs ["key", "text_in"]
   outputs ["text_out"]
   function (SymbFunc "KeyExpansion" 1) "EXP"
   function (SymbFunc "KeyReversal" 1) "REV"
   function (SymbFunc "InitialPermutation" 2) "INIT"
   function (SymbFunc "RoundPermutation" 2) "ROUND"
   function (SymbFunc "FinalPermutation" 2) "FINAL"
   function (SymbFunc "InitialInverse" 2) "INIT_I"
   function (SymbFunc "RoundInverse" 2) "ROUND_I"
   function (SymbFunc "FinalInverse" 2) "FINAL_I"
   register open "R0"
   register open "R1"
   delay open "R2"
   delay open "R3"
   register open "R4"
   joinPoints ["JP", "JP2"]
   connections [("key", "R0_load"), ("R0", "EXP"), ("EXP", "REV"), ("text_in", "R1_load")]
   connectTo_n "EXP" [("INIT", 0), ("ROUND", 0), ("FINAL", 0)]
   connectTo_n "REV" [("INIT_I", 0), ("ROUND_I", 0), ("FINAL_I", 0)]
   connections_n [("R1", ("INIT", 1)), ("INIT", ("JP", 0)), ("JP", ("ROUND", 1))]
   connections_n [("ROUND", ("R2", 0)), ("R2", ("JP", 0)), ("R2", ("FINAL", 1))]
   connections_n [("R1", ("INIT_I", 1)), ("INIT_I", ("JP2", 0)), ("JP2", ("ROUND_I", 1))]
   connections_n [("ROUND_I", ("R3", 0)), ("R3", ("JP2", 0)), ("R3", ("FINAL_I", 1))]
   connections_n [("FINAL", ("R4_load", 0)), ("FINAL_I", ("R4_load", 0)),("R4", ("text_out", 0))]
   constantList ([Var "R2", Var "INIT"] ++ (replicate 10 (Var "R2"))) "TwoList"
   constantList ([Var "R3", Var "INIT_I"] ++ (replicate 10 (Var "R3"))) "ThreeList"
   connect_n ("TwoList_0", ("JP", 1))
   connect_n ("ThreeList_0", ("JP2", 1))

--Configuration for encryption
aes_comb_enc :: Config
aes_comb_enc = (["R0_load", "R1_load", "R4_load"],
               [[Var "key", Var "text_in", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", Var "FINAL"]])

--Configuration for decryption
aes_comb_dec :: Config
aes_comb_dec = (["R0_load", "R1_load", "R4_load"],
               [[Var "key", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
                [Var "R0", Var "text_in", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", open],
	        [Var "R0", Var "R1", Var "FINAL_I"]])

--Demonstrates "Black Box" functions from UsagePattern.hs
testBB :: Datapath
testBB = makeDatapath dp
 where
  dp = do
   inputs ["X", "Y"]
   outputs ["R", "S"]
   funcPipelined "AND_0" AND 4
   funcPipelined "AND_1" AND 3
   connections_n [("X", ("AND_0", 0)), ("X", ("AND_1", 0))]
   connections_n [("Y", ("AND_0", 1)), ("X", ("AND_1", 1))]
   connections [("AND_0_d0", "R"), ("AND_0_d1", "S")]

--BAD EXAMPLES
--Examples of invalid datapaths that demonstrate the checkDatapath function
--All are variants of dp0

bad_NoOutputs :: Datapath
bad_NoOutputs = makeDatapath dp
 where
  dp = do
   input "X"
   function (SymbFunc "f" 1) "F"
   connections [("X", "F"), ("F", "F")]
   registerBefore open "R1" ("F", 0)

bad_notComplete :: Datapath
bad_notComplete = makeDatapath dp
 where
  dp = do
   input "X"
   function (SymbFunc "f" 1) "F"
   output "R"
   connections [("X", "F"), ("X", "R")]
   registerBefore open "R1" ("F", 0)
   registerBefore open "R2" ("R", 0)

bad_Loop :: Datapath
bad_Loop = makeDatapath dp
 where
  dp = do
   input "X"
   function (SymbFunc "f" 1) "F"
   output "R"
   connections [("X", "F"), ("F", "R"), ("F", "F")]
   registerBefore open "R2" ("R", 0)
      
bad_Disjoint :: Datapath
bad_Disjoint = makeDatapath dp
 where
  dp = do
   input "X"
   function (SymbFunc "f" 1) "F"
   output "R"
   connections [("X", "F"), ("F", "F")]
   registerBefore open "R1" ("F", 0)
   registerBefore open "R2" ("R", 0)
