module UsagePattern where

import SymbVal
import Datapath
import Configuration
import STATE


import List

type UsagePattern = ([[Pair]], [[Pair]])

type Latency = Int
type Throughput = (Int, Int)

type UsageEnv = [(SymbVal, NodeName)]

--Creates a pipelined version of a function
--with given latency
funcPipelined :: NodeName -> SymbFunc -> Latency -> Result ()
funcPipelined nm f 0 =
 do
   function f nm
funcPipelined nm f l =
 do
   let nm' = nm ++ "_d"
   function f nm
   delayN l open nm'
   addEdge (nm, (nm' ++ "_" ++ (show (l - 1)), 0))

--Creates an unpipelined version of a function
--with given latency and TP <= 1
funcUnpipelined :: NodeName -> SymbFunc -> Latency -> Throughput -> Result ()
funcUnpipelined nm f l (n, d) =
 do
   let nm' = nm ++ "_d"
   function f nm
   delayN l open nm'
   addEdge (nm, (nm' ++ "_" ++ (show (l - 1)), 0))
   let last = nm' ++ "0"
   addEdge (last, (last,0)) 
   let v1 = replicate n (Var (nm' ++ "1"))
   let v2 = replicate (d - n) (Var last)
   let con = nm ++ "_control"
   constantList (v1 ++ v2) con
   addEdge (con ++ "0", ("FOO", 0))

--An attempt to generate a datapath from an arbitrary usage pattern


bbUsagePattern :: UsagePattern -> Datapath
bbUsagePattern (ins, outs) = makeDatapath res
 where
  outcols = pairsToCols outs
  incols = pairsToCols ins
  res =  buildDP (incols, outcols)

buildDP :: ([Column], [Column]) -> Result ()
buildDP (incols, outcols) =
 do
  let env = concatMap buildEnv incols
  let ins = (map fst incols)
  mapM_ input ins
  mapM_ inputGuard incols
  mapM_ (addPaths env) outcols
  mapM_ outputGuard outcols

buildEnv :: Column -> UsageEnv
buildEnv (node, vs) = map (\v -> (v, node)) vs

addPaths :: UsageEnv -> Column -> Result ()
addPaths env (node, vs) = 
 do
  output node
  let zips = zip vs (range (0, length vs))
  mapM_ (addPath env (node, 0)) zips
  
  
addPath :: UsageEnv -> InputPort -> (SymbVal, Int) -> Result ()
addPath env (node, p) (Unknown, _) = return ()
addPath env (node, p) (FuncApp f vs, n) =
 do
   let func = (node ++ "_" ++ (show n))
   function f func
   connect_n (func, (node, p))
   let k = nParams f
   let zs = zip (replicate k func) (range (0, k))
   let zs' = zip vs (replicate k n)
   let zips = zip zs zs'
   mapM_ (uncurry (addPath env)) zips
addPath env (node, p) (v@(VarIx var ix), n) = --constant or input
 do
   case lookup v env of
     Nothing -> do --constant value
       let c = (node ++ "_" ++ (show n))
       constant v c
       connect_n (c, (node, p))
     Just x -> do
       if n - ix > 0 --latency test
         then do
	   let latname = node ++ "_" ++ var ++ "_" ++ (show ix) ++ "_lat_" ++ (show n)
	   delayN (n - ix) open latname
	   connect (x ++ "_guard", latname ++ "_" ++ (show (n - ix - 1)))
	   connect_n (latname ++ "_0", (node, p))
	 else do
	   connect_n (x ++ "_guard", (node, p))

inputGuard (i, vs) =
 do
   let vs' = map (\x -> if x == Unknown then Unknown else Var i) vs
   joinPoint (i ++ "_guard")
   constantList vs' (i ++ "_glist")
   connect_n (i ++ "_glist_0", (i ++ "_guard",1))
   connect (i, i ++ "_guard")

outputGuard (o, vs) =
 do
   let vs' = makeVals vs 0
   joinPoint (o ++ "_guard")
   constantList vs' (o ++ "_glist")
   connect_n  (o ++ "_glist_0", (o ++ "_guard", 1))
   dp <- readSTATE
   let prvs = getPrevCons dp o
   mapM_ (repoint (o ++ "_guard", 0)) prvs 
   connect (o ++ "_guard", o)
  where
   makeVals [] _ = []
   makeVals (Unknown:vs) n = Unknown : (makeVals vs (n+1))
   makeVals (v:vs) n = (Var (o ++ "_" ++ (show n))) : (makeVals vs (n+1))
--IO Functions
       
displayUP_n :: Int -> UsagePattern -> IO ()
displayUP_n n (ins, outs) = 
 do
   let icols = pairsToCols ins
   let ocols = pairsToCols outs
   let (ihs, idata) = colsToConfig icols
   let (ohs, odata) = colsToConfig ocols
   let headings = ihs ++ ohs
   let bt =  transpose ((transpose idata) ++ (transpose odata))
   let h = concat (intersperse "\t" headings)
   putStr "CC\t|\t"
   putStrLn h
   putStr "--------+"
   putStrLn (replicate (8 * ((length headings)+1)) '-')
   printRows bt 0 n
   return ()
 where
  printRows [] _ _ = return ()
  printRows _ _ 0 = return ()
  printRows (r:rs) c n = 
   do
    putStr (show c)
    putStr "\t|\t"
    putStrLn $ concat (intersperse "\t" (map show r))
    printRows rs (c+1) (n-1)


upToHTML_n :: Int -> UsagePattern -> String
upToHTML_n n (ins, outs) = header ++ (pRows bt 0 n) ++ footer
 where
   icols = pairsToCols ins
   ocols = pairsToCols outs
   (ihs, idata) = colsToConfig icols
   (ohs, odata) = colsToConfig ocols
   headings = ihs ++ ohs
   bt = transpose ((transpose idata) ++ (transpose odata))
   h = concat (intersperse "</TD><TD>" headings)
   header = "<HTML>\n<BODY>\n<TABLE BORDER = \"1\">\n<TR><TD>CC</TD><TD>" ++ h ++ "</TD></TR>\n"
   footer = "</TABLE>\n</BODY>\n</HTML>"
   pRows [] _ _ = []
   pRows _ _ 0 = []
   pRows (r:rs) c n = row ++ (pRows rs (c+1) (n-1))
    where
     row = "<TR><TD>" ++ (show c) ++ ("</TD><TD>") ++ datum ++ "</TD></TR>\n"
     datum = concat (intersperse "</TD><TD>" (map showHTML r))

testUP :: UsagePattern
testUP = ([[("A", VarIx "X" 0), ("B", VarIx "Y" 0)],[("A", open), ("B", open)]], [[("R", FuncApp AND [VarIx "X" 0, VarIx "Y" 0])],[("R", FuncApp OR [VarIx "X" 0, VarIx "Y" 0])]])    

testUP2 :: UsagePattern
testUP2 = ([[("X", VarIx "X" 0), ("Y", VarIx "Y" 0), ("Z", open)],
            [("X", open),        ("Y", VarIx "Y" 1), ("Z", VarIx "Z" 1)],
            [("X", open),        ("Y", open), ("Z", open)]], 
	   [[("R", FuncApp AND [VarIx "X" 0, VarIx "Y" 0])],
	    [("R", FuncApp OR [VarIx "Y" 0, VarIx "Y" 1])],
	    [("R", FuncApp XOR [VarIx "Y" 1, VarIx "Z" 1])]]) 

testUP3 :: UsagePattern
testUP3 = ([[("X", VarIx "X" 0), ("Y", VarIx "Y" 0), ("Z", VarIx "Z" 0)],
            [("X", VarIx "X" 1), ("Y", open),        ("Z", VarIx "Z" 1)],
            [("X", open),        ("Y", VarIx "Y" 2), ("Z", open)],
            [("X", VarIx "X" 3), ("Y", open),        ("Z", open)],
            [("X", open),        ("Y", VarIx "Y" 4), ("Z", open)],
            [("X", open),        ("Y", open),        ("Z", VarIx "Z" 5)]], 
	   [[("R", open), ("S", open)],
	    [("R", FuncApp ADD [VarIx "X" 0, VarIx "Z" 1]), ("S", open)],
	    [("R", FuncApp SUB [VarIx "X" 0, VarIx "Y" 2]), ("S", open)],
	    [("R", FuncApp SUB [VarIx "X" 0, VarIx "X" 1]), ("S", VarIx "X" 3)],
	    [("R", FuncApp SUB [VarIx "X" 0, VarIx "Z" 0]), ("S", VarIx "Y" 4)],
	    [("R", FuncApp SUB [VarIx "X" 0, VarIx "Y" 0]), ("S", FuncApp SUB [VarIx "Z" 5, VarIx "X" 0])]]) 
	       
	       
